#!/usr/bin/env python3

"""
Tool for get model information.
"""

import logging
from typing import Type, Optional, Annotated
from dataclasses import dataclass
import basico
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool
from langchain_core.messages import ToolMessage
from langchain_core.tools.base import InjectedToolCallId
from langgraph.prebuilt import InjectedState
from langgraph.types import Command
from .load_biomodel import ModelData, load_biomodel

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@dataclass
class RequestedModelInfo:
    """
    Dataclass for storing the requested model information.
    """
    species: bool = Field(description="Get species from the model.", default=False)
    parameters: bool = Field(description="Get parameters from the model.", default=False)
    compartments: bool = Field(description="Get compartments from the model.", default=False)
    units: bool = Field(description="Get units from the model.", default=False)
    description: bool = Field(description="Get description from the model.", default=False)
    name: bool = Field(description="Get name from the model.", default=False)

class GetModelInfoInput(BaseModel):
    """
    Input schema for the GetModelInfo tool.
    """
    requested_model_info: RequestedModelInfo = Field(description="requested model information")
    sys_bio_model: ModelData = Field(description="model data")
    tool_call_id: Annotated[str, InjectedToolCallId]
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class GetModelInfoTool(BaseTool):
    """
    This tool ise used extract model information.
    """
    name: str = "get_modelinfo"
    description: str = """A tool for extracting name,
                    description, species, parameters,
                    compartments, and units from a model."""
    args_schema: Type[BaseModel] = GetModelInfoInput

    def _run(self,
            requested_model_info: RequestedModelInfo,
            tool_call_id: Annotated[str, InjectedToolCallId],
            state: Annotated[dict, InjectedState],
            sys_bio_model: Optional[ModelData] = None,
             ) -> Command:
        """
        Run the tool.

        Args:
            requested_model_info (RequestedModelInfo): The requested model information.
            tool_call_id (str): The tool call ID. This is injected by the system.
            state (dict): The state of the tool.
            sys_bio_model (ModelData): The model data.

        Returns:
            Command: The updated state of the tool.
        """
        logger.log(logging.INFO,
                   "Calling get_modelinfo tool %s, %s",
                     sys_bio_model,
                   requested_model_info)
        # print (state, 'state')
        sbml_file_path = state['sbml_file_path'][-1] if len(state['sbml_file_path']) > 0 else None
        model_obj = load_biomodel(sys_bio_model,
                                  sbml_file_path=sbml_file_path)
        dic_results = {}
        # Extract species from the model
        if requested_model_info.species:
            df_species = basico.model_info.get_species(model=model_obj.copasi_model)
            if df_species is None:
                raise ValueError("Unable to extract species from the model.")
            # Convert index into a column
            df_species.reset_index(inplace=True)
            dic_results['Species'] = df_species[
                                        ['name',
                                         'compartment',
                                         'type',
                                         'unit',
                                         'initial_concentration',
                                         'display_name']]
            # Convert this into a dictionary
            dic_results['Species'] = dic_results['Species'].to_dict(orient='records')

        # Extract parameters from the model
        if requested_model_info.parameters:
            df_parameters = basico.model_info.get_parameters(model=model_obj.copasi_model)
            if df_parameters is None:
                raise ValueError("Unable to extract parameters from the model.")
            # Convert index into a column
            df_parameters.reset_index(inplace=True)
            dic_results['Parameters'] = df_parameters[
                                        ['name',
                                         'type',
                                         'unit',
                                         'initial_value',
                                         'display_name']]
            # Convert this into a dictionary
            dic_results['Parameters'] = dic_results['Parameters'].to_dict(orient='records')

        # Extract compartments from the model
        if requested_model_info.compartments:
            df_compartments = basico.model_info.get_compartments(model=model_obj.copasi_model)
            dic_results['Compartments'] = df_compartments.index.tolist()
            dic_results['Compartments'] = ','.join(dic_results['Compartments'])

        # Extract description from the model
        if requested_model_info.description:
            dic_results['Description'] = model_obj.description

        # Extract description from the model
        if requested_model_info.name:
            dic_results['Name'] = model_obj.name

        # Extract time unit from the model
        if requested_model_info.units:
            dic_results['Units'] = basico.model_info.get_model_units(model=model_obj.copasi_model)

        # Prepare the dictionary of updated state for the model
        dic_updated_state_for_model = {}
        for key, value in {
                        "model_id": [sys_bio_model.biomodel_id],
                        "sbml_file_path": [sbml_file_path],
                        }.items():
            if value:
                dic_updated_state_for_model[key] = value

        return Command(
            update=dic_updated_state_for_model|{
                    # update the message history
                    "messages": [
                        ToolMessage(
                            content=dic_results,
                            tool_call_id=tool_call_id
                            )
                        ],
                    }
            )



"""
Test cases for utils/enrichments/enrichments.py
"""

from ..utils.enrichments.enrichments import Enrichments

class TestEnrichments(Enrichments):
    """Test implementation of the Enrichments interface for testing purposes."""

    def enrich_documents(self, texts: list[str]) -> list[list[float]]:
        return [
            f"Additional text description of {text} as the input." for text in texts
        ]

    def enrich_documents_with_rag(self, texts, docs):
        # Currently we don't have a RAG model to test this method.
        # Thus, we will just call the enrich_documents method instead.
        return self.enrich_documents(texts)

def test_enrich_documents():
    """Test enriching documents using the Enrichments interface."""
    enrichments = TestEnrichments()
    texts = ["text1", "text2"]
    result = enrichments.enrich_documents(texts)
    assert result == [
        "Additional text description of text1 as the input.",
        "Additional text description of text2 as the input.",
    ]

def test_enrich_documents_with_rag():
    """Test enriching documents with RAG using the Enrichments interface."""
    enrichments = TestEnrichments()
    texts = ["text1", "text2"]
    docs = ["doc1", "doc2"]
    result = enrichments.enrich_documents_with_rag(texts, docs)
    assert result == [
        "Additional text description of text1 as the input.",
        "Additional text description of text2 as the input.",
    ]



"""
Test cases for utils/kg_utils.py
"""

import pytest
import networkx as nx
import pandas as pd
from ..utils import kg_utils


@pytest.fixture(name="sample_graph")
def make_sample_graph():
    """Return a sample graph"""
    sg = nx.Graph()
    sg.add_node(1, node_id=1, feature_id="A", feature_value="ValueA")
    sg.add_node(2, node_id=2, feature_id="B", feature_value="ValueB")
    sg.add_edge(1, 2, edge_id=1, feature_id="E", feature_value="EdgeValue")
    return sg


def test_kg_to_df_pandas(sample_graph):
    """Test the kg_to_df_pandas function"""
    df_nodes, df_edges = kg_utils.kg_to_df_pandas(sample_graph)
    print(df_nodes)
    expected_nodes_data = {
        "node_id": [1, 2],
        "feature_id": ["A", "B"],
        "feature_value": ["ValueA", "ValueB"],
    }
    expected_nodes_df = pd.DataFrame(expected_nodes_data, index=[1, 2])
    print(expected_nodes_df)
    expected_edges_data = {
        "node_source": [1],
        "node_target": [2],
        "edge_id": [1],
        "feature_id": ["E"],
        "feature_value": ["EdgeValue"],
    }
    expected_edges_df = pd.DataFrame(expected_edges_data)

    # Assert that the dataframes are equal but the order of columns may be different
    # Ignore the index of the dataframes
    pd.testing.assert_frame_equal(df_nodes, expected_nodes_df, check_like=True)
    pd.testing.assert_frame_equal(df_edges, expected_edges_df, check_like=True)


def test_df_pandas_to_kg():
    """Test the df_pandas_to_kg function"""
    nodes_data = {
        "node_id": [1, 2],
        "feature_id": ["A", "B"],
        "feature_value": ["ValueA", "ValueB"],
    }
    df_nodes_attrs = pd.DataFrame(nodes_data).set_index("node_id")

    edges_data = {
        "node_source": [1],
        "node_target": [2],
        "edge_id": [1],
        "feature_id": ["E"],
        "feature_value": ["EdgeValue"],
    }
    df_edges = pd.DataFrame(edges_data)

    kg = kg_utils.df_pandas_to_kg(
        df_edges, df_nodes_attrs, "node_source", "node_target"
    )

    assert len(kg.nodes) == 2
    assert len(kg.edges) == 1

    assert kg.nodes[1]["feature_id"] == "A"
    assert kg.nodes[1]["feature_value"] == "ValueA"
    assert kg.nodes[2]["feature_id"] == "B"
    assert kg.nodes[2]["feature_value"] == "ValueB"

    assert kg.edges[1, 2]["feature_id"] == "E"
    assert kg.edges[1, 2]["feature_value"] == "EdgeValue"
    assert kg.edges[1, 2]["edge_id"] == 1



#!/usr/bin/env python3

"""
Abstract class for dataset.
"""

from abc import ABC, abstractmethod

class Dataset(ABC):
    """
    Abstract class for dataset.
    """
    @abstractmethod
    def setup(self):
        """
        A method to set up the dataset.
        """

    @abstractmethod
    def load_data(self):
        """
        A method to load the dataset and potentially preprocess it.
        """



#!/usr/bin/env python3

"""
multi_paper_rec: Tool for getting recommendations
                based on multiple papers
"""

import json
import logging
from typing import Annotated, Any, List, Optional
import hydra
import requests
from langchain_core.messages import ToolMessage
from langchain_core.tools import tool
from langchain_core.tools.base import InjectedToolCallId
from langgraph.types import Command
from pydantic import BaseModel, Field


# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class MultiPaperRecInput(BaseModel):
    """Input schema for multiple paper recommendations tool."""

    paper_ids: List[str] = Field(
        description=("List of Semantic Scholar Paper IDs to get recommendations for")
    )
    limit: int = Field(
        default=2,
        description="Maximum total number of recommendations to return",
        ge=1,
        le=500,
    )
    year: Optional[str] = Field(
        default=None,
        description="Year range in format: YYYY for specific year, "
        "YYYY- for papers after year, -YYYY for papers before year, or YYYY:YYYY for range",
    )
    tool_call_id: Annotated[str, InjectedToolCallId]

    model_config = {"arbitrary_types_allowed": True}


# Load hydra configuration
with hydra.initialize(version_base=None, config_path="../../configs"):
    cfg = hydra.compose(
        config_name="config", overrides=["tools/multi_paper_recommendation=default"]
    )
    cfg = cfg.tools.multi_paper_recommendation


@tool(args_schema=MultiPaperRecInput, parse_docstring=True)
def get_multi_paper_recommendations(
    paper_ids: List[str],
    tool_call_id: Annotated[str, InjectedToolCallId],
    limit: int = 2,
    year: Optional[str] = None,
) -> Command[Any]:
    """
    Get recommendations for a group of multiple papers using the Semantic Scholar IDs.
    No other paper IDs are supported.

    Args:
        paper_ids (List[str]): The list of paper IDs to base recommendations on.
        tool_call_id (Annotated[str, InjectedToolCallId]): The tool call ID.
        limit (int, optional): The maximum number of recommendations to return. Defaults to 2.
        year (str, optional): Year range for papers.
        Supports formats like "2024-", "-2024", "2024:2025". Defaults to None.

    Returns:
        Dict[str, Any]: The recommendations and related information.
    """
    logging.info(
        "Starting multi-paper recommendations search with paper IDs: %s", paper_ids
    )

    endpoint = cfg.api_endpoint
    headers = cfg.headers
    payload = {"positivePaperIds": paper_ids, "negativePaperIds": []}
    params = {
        "limit": min(limit, 500),
        "fields": ",".join(cfg.api_fields),
    }

    # Add year parameter if provided
    if year:
        params["year"] = year

    # Getting recommendations
    response = requests.post(
        endpoint,
        headers=headers,
        params=params,
        data=json.dumps(payload),
        timeout=cfg.request_timeout,
    )
    logging.info(
        "API Response Status for multi-paper recommendations: %s", response.status_code
    )

    data = response.json()
    recommendations = data.get("recommendedPapers", [])

    if not recommendations:
        return Command(
            update={  # Place 'messages' inside 'update'
                "messages": [
                    ToolMessage(
                        content="No recommendations found based on multiple papers.",
                        tool_call_id=tool_call_id,
                    )
                ]
            }
        )

    # Create a dictionary to store the papers
    filtered_papers = {
        paper["paperId"]: {
            # "semantic_scholar_id": paper["paperId"],  # Store Semantic Scholar ID
            "Title": paper.get("title", "N/A"),
            "Abstract": paper.get("abstract", "N/A"),
            "Year": paper.get("year", "N/A"),
            "Citation Count": paper.get("citationCount", "N/A"),
            "URL": paper.get("url", "N/A"),
            # "arXiv_ID": paper.get("externalIds", {}).get(
            #     "ArXiv", "N/A"
            # ),  # Extract arXiv ID
        }
        for paper in recommendations
        if paper.get("title") and paper.get("authors")
    }

    content = "Recommendations based on multiple papers was successful."
    content += " Here is a summary of the recommendations:"
    content += f"Number of papers found: {len(filtered_papers)}\n"
    content += f"Query Paper IDs: {', '.join(paper_ids)}\n"
    content += f"Year: {year}\n" if year else ""

    return Command(
        update={
            "multi_papers": filtered_papers,  # Now sending the dictionary directly
            "last_displayed_papers": "multi_papers",
            "messages": [
                ToolMessage(
                    content=content,
                    tool_call_id=tool_call_id,
                    artifact=filtered_papers,
                )
            ],
        }
    )



"""
Test cases for datasets/dataset.py
"""

from ..datasets.dataset import Dataset

class MockDataset(Dataset):
    """
    Mock dataset class for testing purposes.
    """
    def setup(self):
        pass

    def load_data(self):
        pass

def test_dataset_setup():
    """
    Test the setup method of the Dataset class.
    """
    dataset = MockDataset()
    assert dataset.setup() is None

def test_dataset_load_data():
    """
    Test the load_data method of the Dataset class.
    """
    dataset = MockDataset()
    assert dataset.load_data() is None



"""
Integration tests for talk2scholars system with OpenAI.
"""

import os
import pytest
import hydra
from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage, AIMessage
from ..agents.main_agent import get_app
from ..state.state_talk2scholars import Talk2Scholars

# pylint: disable=redefined-outer-name


@pytest.mark.skipif(
    not os.getenv("OPENAI_API_KEY"), reason="Requires OpenAI API key to run"
)
def test_main_agent_real_llm():
    """
    Test that the main agent invokes S2 agent correctly
    and updates the state with real LLM execution.
    """

    # Load Hydra Configuration EXACTLY like in main_agent.py
    with hydra.initialize(version_base=None, config_path="../configs"):
        cfg = hydra.compose(
            config_name="config", overrides=["agents/talk2scholars/main_agent=default"]
        )
    hydra_cfg = cfg.agents.talk2scholars.main_agent

    assert hydra_cfg is not None, "Hydra config failed to load"

    # Use the real OpenAI API (ensure env variable is set)
    llm = ChatOpenAI(model="gpt-4o-mini", temperature=hydra_cfg.temperature)

    # Initialize main agent workflow (WITH real Hydra config)
    thread_id = "test_thread"
    app = get_app(thread_id, llm)

    # Provide an actual user query
    initial_state = Talk2Scholars(
        messages=[HumanMessage(content="Find AI papers on transformers")]
    )

    # Invoke the agent (triggers supervisor → s2_agent)
    result = app.invoke(
        initial_state,
        {"configurable": {"config_id": thread_id, "thread_id": thread_id}},
    )

    # Assert that the supervisor routed correctly
    assert "messages" in result, "Expected messages in response"

    # Fix: Accept AIMessage as a valid response type
    assert isinstance(
        result["messages"][-1], (HumanMessage, AIMessage, str)
    ), "Last message should be a valid response"



'''
This file contains the unit tests for the BioModel class.
'''

from typing import List, Dict, Union, Optional
from pydantic import Field
import pytest
from ..models.sys_bio_model import SysBioModel

class TestBioModel(SysBioModel):
    '''
    A test BioModel class for unit testing.
    '''

    biomodel_id: Optional[int] = Field(None, description="BioModel ID of the model")
    sbml_file_path: Optional[str] = Field(None, description="Path to an SBML file")
    name: Optional[str] = Field(..., description="Name of the model")
    description: Optional[str] = Field("", description="Description of the model")
    param1: Optional[float] = Field(0.0, description="Parameter 1")
    param2: Optional[float] = Field(0.0, description="Parameter 2")

    def get_model_metadata(self) -> Dict[str, Union[str, int]]:
        '''
        Get the metadata of the model.
        '''
        return self.biomodel_id

    def update_parameters(self, parameters):
        '''
        Update the model parameters.
        '''
        self.param1 = parameters.get('param1', 0.0)
        self.param2 = parameters.get('param2', 0.0)

    def simulate(self, duration: Union[int, float]) -> List[float]:
        '''
        Simulate the model.
        '''
        return [self.param1 + self.param2 * t for t in range(int(duration))]

def test_get_model_metadata():
    '''
    Test the get_model_metadata method of the BioModel class.
    '''
    model = TestBioModel(biomodel_id=123, name="Test Model", description="A test model")
    metadata = model.get_model_metadata()
    assert metadata == 123

def test_check_biomodel_id_or_sbml_file_path():
    '''
    Test the check_biomodel_id_or_sbml_file_path method of the BioModel class.
    '''
    with pytest.raises(ValueError):
        TestBioModel(name="Test Model", description="A test model")

def test_simulate():
    '''
    Test the simulate method of the BioModel class.
    '''
    model = TestBioModel(biomodel_id=123, name="Test Model", description="A test model")
    model.update_parameters({'param1': 1.0, 'param2': 2.0})
    results = model.simulate(duration=4.0)
    assert results == [1.0, 3.0, 5.0, 7.0]



'''
An abstract base class for BioModels in the BioModels repository.
'''

from abc import ABC, abstractmethod
from typing import List, Dict, Union, Optional
from pydantic import BaseModel, Field, model_validator

class SysBioModel(ABC, BaseModel):
    """
    Abstract base class for BioModels in the BioModels repository.
    This class serves as a general structure for models, allowing
    different mathematical approaches to be implemented in subclasses.
    """
    biomodel_id: Optional[int] = Field(None, description="BioModel ID of the model")
    sbml_file_path: Optional[str] = Field(None, description="Path to an SBML file")
    name: Optional[str] = Field(..., description="Name of the model")
    description: Optional[str] = Field("", description="Description of the model")

    @model_validator(mode="after")
    def check_biomodel_id_or_sbml_file_path(self):
        """
        Validate that either biomodel_id or sbml_file_path is provided.
        """
        if not self.biomodel_id and not self.sbml_file_path:
            raise ValueError("Either biomodel_id or sbml_file_path must be provided.")
        return self

    @abstractmethod
    def get_model_metadata(self) -> Dict[str, Union[str, int]]:
        """
        Abstract method to retrieve metadata of the model.
        This method should return a dictionary containing model metadata.

        Returns:
            dict: Dictionary with model metadata
        """
    @abstractmethod
    def update_parameters(self, parameters: Dict[str, Union[float, int]]) -> None:
        """
        Abstract method to update model parameters.

        Args:
            parameters: Dictionary of parameter values.
        """

    @abstractmethod
    def simulate(self, duration: Union[int, float]) -> List[float]:
        """
        Abstract method to run a simulation of the model.

        Args:
            duration: Duration of the simulation.

        Returns:
            list: List of simulation results.
        """



#!/usr/bin/env python3

"""
A utility module for defining the dataclasses
for the arguments to set up initial settings
before the experiment is run.
"""

import logging
from dataclasses import dataclass
from typing import Union, List, Optional, Annotated
from pydantic import Field
import basico

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@dataclass
class TimeData:
    """
    Dataclass for storing the time data.
    """
    duration: Union[int, float] = Field(
        description="Duration of the simulation",
        default=100)
    interval: Union[int, float] = Field(
        description="The interval is the time step or"
        " the step size of the simulation. It is unrelated"
        " to the step size of species concentration and parameter values.",
        default=100)

@dataclass
class SpeciesInitialData:
    """
    Dataclass for storing the species initial data.
    """
    species_name: List[str] = Field(
        description="List of species whose initial concentration is to be set."
        " This does not include species that reoccur or the species whose"
        " concentration is to be determined/observed at the end of the experiment."
        " Do not hallucinate the species name.",
        default=[])
    species_concentration: List[Union[int, float]] = Field(
        description="List of initial concentrations of species."
        " This does not include species that reoccur or the species whose"
        " concentration is to be determined/observed at the end of the experiment."
        " Do not hallucinate the species concentration.",
        default=[])

@dataclass
class TimeSpeciesNameConcentration:
    """
    Dataclass for storing the time,
    species name, and concentration data.
    """
    time: Union[int, float] = Field(description="time point where the event occurs")
    species_name: str = Field(description="species name")
    species_concentration: Union[int, float] = Field(
        description="species concentration at the time point")

@dataclass
class ReocurringData:
    """
    Dataclass for species that reoccur. In other words,
    the concentration of the species resets to a certain
    value after a certain time interval.
    """
    data: List[TimeSpeciesNameConcentration] = Field(
        description="List of time, name, and concentration data"
                    " of species or parameters that reoccur",
                    default=[])

@dataclass
class ArgumentData:
    """
    Dataclass for storing the argument data.
    """
    experiment_name: Annotated[str, "An AI assigned _ separated name of"
                                    " the experiment based on human query"
                                    " and the context of the experiment."
                                    " This must be set before the experiment is run."]
    time_data: Optional[TimeData] = Field(
        description="time data",
        default=None)
    species_to_be_analyzed_before_experiment: Optional[SpeciesInitialData] = Field(
        description="Data of species whose initial concentration"
        " is to be set before the experiment. This does not include"
        " species that reoccur or the species whose concentration"
        " is to be determined at the end of the experiment.",
        default=None)
    reocurring_data: Optional[ReocurringData] = Field(
        description="List of concentration and time data of species that"
        " reoccur. For example, a species whose concentration resets"
        " to a certain value after a certain time interval.",
        default=None)

def add_rec_events(model_object, reocurring_data):
    """
    Add reocurring events to the model.

    Args:
        model_object: The model object.
        reocurring_data: The reocurring data.

    Returns:
        None
    """
    for row in reocurring_data.data:
        tp, sn, sc = row.time, row.species_name, row.species_concentration
        basico.add_event(f'{sn}_{tp}',
                            f'Time > {tp}',
                            [[sn, str(sc)]],
                            model=model_object.copasi_model)



"""
Enrichments interface
"""

from abc import ABC, abstractmethod

class Enrichments(ABC):
    """Interface for enrichment models.

    This is an interface meant for implementing text enrichment models.

    Enrichment models are used to enrich node or relation features in a given knowledge graph.
    """

    @abstractmethod
    def enrich_documents(self, texts: list[str]) -> list[list[str]]:
        """Enrich documents.

        Args:
            texts: List of documents to enrich.

        Returns:
            List of enriched documents.
        """

    @abstractmethod
    def enrich_documents_with_rag(self, texts: list[str], docs: list[str]) -> list[str]:
        """Enrich documents with RAG.

        Args:
            texts: List of documents to enrich.
            docs: List of reference documents to enrich the input texts.

        Returns:
            List of enriched documents with RAG.
        """



#!/usr/bin/env python3

'''
This is the state file for the Talk2Cells agent.
'''

from langgraph.prebuilt.chat_agent_executor import AgentState

class Talk2Cells(AgentState):
    """
    The state for the Talk2Cells agent.
    """
    search_table: str



"""
Test cases for utils/embeddings/sentence_transformer.py
"""

import pytest
import numpy as np
from ..utils.embeddings.sentence_transformer import EmbeddingWithSentenceTransformer

@pytest.fixture(name="embedding_model")
def embedding_model_fixture():
    """
    Fixture for creating an instance of EmbeddingWithSentenceTransformer.
    """
    model_name = "sentence-transformers/all-MiniLM-L6-v1"  # Small model for testing
    return EmbeddingWithSentenceTransformer(model_name=model_name)

def test_embed_documents(embedding_model):
    """
    Test the embed_documents method of EmbeddingWithSentenceTransformer class.
    """
    # Perform embedding
    texts = ["This is a test sentence.", "Another test sentence."]
    embeddings = embedding_model.embed_documents(texts)
    # Check the result
    assert len(embeddings) == len(texts)
    assert len(embeddings[0]) > 0
    assert len(embeddings[0]) == 384
    assert embeddings.dtype == np.float32

def test_embed_query(embedding_model):
    """
    Test the embed_query method of EmbeddingWithSentenceTransformer class.
    """
    # Perform embedding
    text = "This is a test query."
    embedding = embedding_model.embed_query(text)
    # Check the result
    assert len(embedding) > 0
    assert len(embedding) == 384
    assert embedding.dtype == np.float32



'''
Test cases for Talk2Biomodels get_annotation tool.
'''

import random
import pytest
from langchain_core.messages import HumanMessage, ToolMessage
from langchain_openai import ChatOpenAI
from ..agents.t2b_agent import get_app
from ..tools.get_annotation import prepare_content_msg

@pytest.fixture(name="make_graph")
def make_graph_fixture():
    '''
    Create an instance of the talk2biomodels agent.
    '''
    unique_id = random.randint(1000, 9999)
    graph = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}
    graph.update_state(config, {"llm_model": ChatOpenAI(model='gpt-4o-mini',
                                    temperature=0)
                                })
    return graph, config

def test_no_model_provided(make_graph):
    '''
    Test the tool by not specifying any model.
    We are testing a condition where the user
    asks for annotations of all species without
    specifying a model.
    '''
    app, config = make_graph
    prompt = "Extract annotations of all species. Call the tool get_annotation."
    app.invoke({"messages": [HumanMessage(content=prompt)]},
                        config=config
                    )
    current_state = app.get_state(config)
    # Assert that the state key model_id is empty.
    assert current_state.values["model_id"] == []

def test_valid_species_provided(make_graph):
    '''
    Test the tool by providing a specific species name.
    We are testing a condition where the user asks for annotations
    of a specific species in a specific model.
    '''
    # Test with a valid species name
    app, config = make_graph
    prompt = "Extract annotations of species IL6 in model 537."
    app.invoke(
                {"messages": [HumanMessage(content=prompt)]},
                config=config
            )
    current_state = app.get_state(config)
    # print (current_state.values["dic_annotations_data"])
    dic_annotations_data = current_state.values["dic_annotations_data"]

    # The assert statement checks if IL6 is present in the returned annotations.
    assert dic_annotations_data[0]['data']["Species Name"][0] == "IL6"

def test_invalid_species_provided(make_graph):
    '''
    Test the tool by providing an invalid species name.
    We are testing a condition where the user asks for annotations
    of an invalid species in a specific model.
    '''
    # Test with an invalid species name
    app, config = make_graph
    prompt = "Extract annotations of only species NADH in model 537."
    app.invoke(
        {"messages": [HumanMessage(content=prompt)]},
        config=config
    )
    current_state = app.get_state(config)
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages until a
    # ToolMessage is found.

    test_condition = False
    for msg in reversed_messages:
        # Assert that the one of the messages is a ToolMessage
        # and its artifact is None.
        if isinstance(msg, ToolMessage) and msg.name == "get_annotation":
            #If a ToolMessage exists and artifact is None (meaning no valid annotation was found)
            #and the rejected species (NADH) is mentioned, the test passes.
            if msg.artifact is None and msg.status == "error":
                #If artifact is None, it means no annotation was found
                # (likely due to an invalid species).
                test_condition = True
                break
    assert test_condition

def test_invalid_and_valid_species_provided(make_graph):
    '''
    Test the tool by providing an invalid species name and a valid species name.
    We are testing a condition where the user asks for annotations
    of an invalid species and a valid species in a specific model.
    '''
    # Test with an invalid species name and a valid species name
    app, config = make_graph
    prompt = "Extract annotations of species NADH, NAD, and IL7 in model 64."
    app.invoke(
        {"messages": [HumanMessage(content=prompt)]},
        config=config
    )
    current_state = app.get_state(config)
    dic_annotations_data = current_state.values["dic_annotations_data"]
    # List of species that are expected to be found in the annotations
    extracted_species = []
    for idx in dic_annotations_data[0]['data']["Species Name"]:
        extracted_species.append(dic_annotations_data[0]['data']["Species Name"][idx])
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages until a
    # ToolMessage is found.
    tool_status_success = False
    for msg in reversed_messages:
        # Assert that the one of the messages is a ToolMessage
        # and its artifact is None.
        if isinstance(msg, ToolMessage) and msg.name == "get_annotation":
            if msg.artifact is True and msg.status == "success":
                tool_status_success = True
                break
    assert tool_status_success
    assert set(extracted_species) == set(["NADH", "NAD"])

def test_all_species_annotations(make_graph):
    '''
    Test the tool by asking for annotations of all species is specific models.
    Here, we test the tool with three models since they have different use cases:
        - model 12 contains a species with no URL provided.
        - model 20 contains a species without description.
        - model 56 contains a species with database outside of KEGG, UniProt, and OLS.

    We are testing a condition where the user asks for annotations
    of all species in a specific model.
    '''
    # Loop through the models and test the tool
    # for each model's unique use case.
    for model_id in [12, 20, 56]:
        app, config = make_graph
        prompt = f"Extract annotations of all species model {model_id}."
        # Test the tool get_modelinfo
        app.invoke({"messages": [HumanMessage(content=prompt)]},
                            config=config
                        )
        current_state = app.get_state(config)

        reversed_messages = current_state.values["messages"][::-1]
        # Coveres all of the use cases for the expecetd sting on all the species
        test_condition = False
        for msg in reversed_messages:
            # Skip messages that are not ToolMessages and those that are not
            # from the get_annotation tool.
            if not isinstance(msg, ToolMessage) or msg.name != "get_annotation":
                continue
            if model_id == 12:
                # Extact the first and second description of the LacI protein
                # We already know that the first or second description is missing ('-')
                dic_annotations_data = current_state.values["dic_annotations_data"][0]
                first_descp_laci_protein = dic_annotations_data['data']['Description'][0]
                second_descp_laci_protein = dic_annotations_data['data']['Description'][1]

                # Expect a successful extraction (artifact is True) and that the content
                # matches what is returned by prepare_content_msg for species.
                # And that the first or second description of the LacI protein is missing.
                if (msg.artifact is True and msg.content == prepare_content_msg([])
                    and msg.status=="success" and (first_descp_laci_protein == '-' or
                                                    second_descp_laci_protein == '-')):
                    test_condition = True
                    break

            if model_id == 20:
                # Expect an error message containing a note
                # that species extraction failed.
                if ("Unable to extract species from the model"
                    in msg.content and msg.status == "error"):
                    test_condition = True
                    break

            if model_id == 56:
                # Expect a successful extraction (artifact is True) and that the content
                # matches for for missing description ['ORI'].
                if (msg.artifact is True and
                msg.content == prepare_content_msg(['ORI'])
                and msg.status == "success"):
                    test_condition = True
                    break
        assert test_condition # Expected output is validated



'''
A test BasicoModel class for pytest unit testing.
'''

import pandas as pd
import pytest
import basico
from ..models.basico_model import BasicoModel

@pytest.fixture(name="model")
def model_fixture():
    """
    A fixture for the BasicoModel class.
    """
    return BasicoModel(biomodel_id=64, species={"Pyruvate": 100}, duration=2, interval=2)

def test_with_biomodel_id(model):
    """
    Test initialization of BasicoModel with biomodel_id.
    """
    assert model.biomodel_id == 64
    model.update_parameters(parameters={'Pyruvate': 0.5, 'KmPFKF6P': 1.5})
    df_species = basico.model_info.get_species(model=model.copasi_model)
    assert df_species.loc['Pyruvate', 'initial_concentration'] == 0.5
    df_parameters = basico.model_info.get_parameters(model=model.copasi_model)
    assert df_parameters.loc['KmPFKF6P', 'initial_value'] == 1.5
    # check if the simulation results are a pandas DataFrame object
    assert isinstance(model.simulate(duration=2, interval=2), pd.DataFrame)
    # Pass a None value to the update_parameters method
    # and it should not raise an error
    model.update_parameters(parameters={None: None})
    # check if the model description is updated
    assert model.description == basico.biomodels.get_model_info(model.biomodel_id)["description"]
    # check if an error is raised if an invalid species/parameter (`Pyruv`)
    # is passed and it should raise a ValueError
    with pytest.raises(ValueError):
        model.update_parameters(parameters={'Pyruv': 0.5})

def test_with_sbml_file():
    """
    Test initialization of BasicoModel with sbml_file_path.
    """
    model_object = BasicoModel(sbml_file_path="./BIOMD0000000064_url.xml")
    assert model_object.sbml_file_path == "./BIOMD0000000064_url.xml"
    assert isinstance(model_object.simulate(duration=2, interval=2), pd.DataFrame)

def test_check_biomodel_id_or_sbml_file_path():
    '''
    Test the check_biomodel_id_or_sbml_file_path method of the BioModel class.
    '''
    with pytest.raises(ValueError):
        BasicoModel(species={"Pyruvate": 100}, duration=2, interval=2)

def test_get_model_metadata():
    """
    Test the get_model_metadata method of the BasicoModel class.
    """
    model = BasicoModel(biomodel_id=64)
    metadata = model.get_model_metadata()
    assert metadata["Model Type"] == "SBML Model (COPASI)"
    assert metadata["Parameter Count"] == len(basico.get_parameters())



"""
This module contains the API for fetching Kegg database
"""
import re
from typing import List, Dict
import requests

def fetch_from_api(base_url: str, query: str) -> str:
    """Fetch data from the given API endpoint."""
    try:
        response = requests.get(base_url + query, timeout=10)
        response.raise_for_status()
        return response.text
    except requests.exceptions.RequestException as e:
        print(f"Error fetching data for query {query}: {e}")
        return ""

def fetch_kegg_names(ids: List[str], batch_size: int = 10) -> Dict[str, str]:
    """
    Fetch the names of multiple KEGG entries using the KEGG REST API in batches.

    Args:
        ids (List[str]): List of KEGG IDs.
        batch_size (int): Maximum number of IDs to include in a single request.

    Returns:
        Dict[str, str]: A mapping of KEGG IDs to their names.
    """
    if not ids:
        return {}

    base_url = "https://rest.kegg.jp/get/"
    entry_name_map = {}

    # Process IDs in batches
    for i in range(0, len(ids), batch_size):
        batch = ids[i:i + batch_size]
        query = "+".join(batch)
        entry_data = fetch_from_api(base_url, query)

        # if not entry_data:
        #     continue
        entries = entry_data.split("///")
        for entry in entries:
            if not entry.strip():
                continue
            lines = entry.strip().split("\n")
            entry_line = next((line for line in lines
                                if line.startswith("ENTRY")), None)
            name_line = next((line for line in lines
                                if line.startswith("NAME")), None)

            # if not entry_line and not name_line:
            #     continue
            entry_id = entry_line.split()[1]
            # Split multiple names in the NAME field and clean them
            names = [
                re.sub(r'[^a-zA-Z0-9\s]', '', name).strip()
                for name in name_line.replace("NAME", "").strip().split(";")
            ]
            # Join cleaned names into a single string
            entry_name_map[entry_id] = " ".join(names).strip()

    return entry_name_map

def fetch_kegg_annotations(data: List[Dict[str, str]],
                           batch_size: int = 10) -> Dict[str, Dict[str, str]]:
    """Fetch KEGG entry descriptions grouped by database type."""
    grouped_data = {}
    for entry in data:
        db_type = entry["Database"].lower()
        grouped_data.setdefault(db_type, []).append(entry["Id"])

    results = {}
    for db_type, ids in grouped_data.items():
        results[db_type] = fetch_kegg_names(ids, batch_size=batch_size)

    return results

# def get_protein_name_or_label(data: List[Dict[str, str]],
#                               batch_size: int = 10) -> Dict[str, Dict[str, str]]:
#     """Fetch descriptions for KEGG-related identifiers."""
#     return fetch_kegg_annotations(data, batch_size=batch_size)



#!/usr/bin/env python3

"""
Main agent for the talk2scholars app using ReAct pattern.

This module implements a hierarchical agent system where a supervisor agent
routes queries to specialized sub-agents. It follows the LangGraph patterns
for multi-agent systems and implements proper state management.
"""

import logging
from typing import Literal, Callable
from pydantic import BaseModel
import hydra
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import SystemMessage, HumanMessage, AIMessage
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import END, START, StateGraph
from langgraph.types import Command
from ..agents import s2_agent
from ..state.state_talk2scholars import Talk2Scholars

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_hydra_config():
    """
    Loads the Hydra configuration for the main agent.

    This function initializes the Hydra configuration system and retrieves the settings
    for the `Talk2Scholars` agent, ensuring that all required parameters are loaded.

    Returns:
        DictConfig: The configuration object containing parameters for the main agent.
    """
    with hydra.initialize(version_base=None, config_path="../configs"):
        cfg = hydra.compose(
            config_name="config", overrides=["agents/talk2scholars/main_agent=default"]
        )
    return cfg.agents.talk2scholars.main_agent


def make_supervisor_node(llm_model: BaseChatModel, thread_id: str) -> Callable:
    """
    Creates the supervisor node responsible for routing user queries to the appropriate sub-agents.

    This function initializes the routing logic by leveraging the system and router prompts defined
    in the Hydra configuration. The supervisor determines whether to
    call a sub-agent (like `s2_agent`)
    or directly generate a response using the language model.

    Args:
        llm_model (BaseChatModel): The language model used for decision-making.
        thread_id (str): Unique identifier for the current conversation session.

    Returns:
        Callable: The supervisor node function that processes user queries and
        decides the next step.
    """
    cfg = get_hydra_config()
    logger.info("Hydra configuration for Talk2Scholars main agent loaded: %s", cfg)
    members = ["s2_agent"]
    options = ["FINISH"] + members
    # Define system prompt for general interactions
    system_prompt = cfg.system_prompt
    # Define router prompt for routing to sub-agents
    router_prompt = cfg.router_prompt

    class Router(BaseModel):
        """Worker to route to next. If no workers needed, route to FINISH."""

        next: Literal[*options]

    def supervisor_node(
        state: Talk2Scholars,
    ) -> Command:
        """
        Handles the routing logic for the supervisor agent.

        This function determines the next agent to invoke based on the router prompt response.
        If no further processing is required, it generates an AI response using the system prompt.

        Args:
            state (Talk2Scholars): The current conversation state, including messages
            exchanged so far.

        Returns:
            Command: A command dictating whether to invoke a sub-agent or generate a final response.
        """
        messages = [SystemMessage(content=router_prompt)] + state["messages"]
        structured_llm = llm_model.with_structured_output(Router)
        response = structured_llm.invoke(messages)
        goto = response.next
        logger.info("Routing to: %s, Thread ID: %s", goto, thread_id)
        if goto == "FINISH":
            goto = END  # Using END from langgraph.graph
            # If no agents were called, and the last message was
            # from the user, call the LLM to respond to the user
            # with a slightly different system prompt.
            if isinstance(messages[-1], HumanMessage):
                response = llm_model.invoke(
                    [
                        SystemMessage(content=system_prompt),
                    ]
                    + messages[1:]
                )
                return Command(
                    goto=goto, update={"messages": AIMessage(content=response.content)}
                )
        # Go to the requested agent
        return Command(goto=goto)

    return supervisor_node


def get_app(
    thread_id: str,
    llm_model: BaseChatModel = ChatOpenAI(model="gpt-4o-mini", temperature=0),
):
    """
    Initializes and returns the LangGraph-based hierarchical agent system.

    This function constructs the agent workflow by defining nodes for the supervisor
    and sub-agents. It compiles the graph using `StateGraph` to enable structured
    conversational workflows.

    Args:
        thread_id (str): A unique session identifier for tracking conversation state.
        llm_model (BaseChatModel, optional): The language model used for query processing.
            Defaults to `ChatOpenAI(model="gpt-4o-mini", temperature=0)`.

    Returns:
        StateGraph: A compiled LangGraph application that can process user queries.

    Example:
        >>> app = get_app("thread_123")
        >>> result = app.invoke(initial_state)
    """
    cfg = get_hydra_config()

    def call_s2_agent(
        state: Talk2Scholars,
    ) -> Command[Literal["supervisor"]]:
        """
        Invokes the Semantic Scholar (S2) agent to retrieve relevant research papers.

        This function calls the `s2_agent` and updates the conversation state with retrieved
        academic papers. The agent uses Semantic Scholar's API to find papers based on
        user queries.

        Args:
            state (Talk2Scholars): The current state of the conversation, containing messages
                and any previous search results.

        Returns:
            Command: A command to update the conversation state with the retrieved papers
                and return control to the supervisor node.

        Example:
            >>> result = call_s2_agent(current_state)
            >>> next_step = result.goto
        """
        logger.info("Calling S2 agent")
        app = s2_agent.get_app(thread_id, llm_model)

        # Invoke the S2 agent, passing state,
        # Pass both config_id and thread_id
        response = app.invoke(
            state,
            {
                "configurable": {
                    "config_id": thread_id,
                    "thread_id": thread_id,
                }
            },
        )
        logger.info("S2 agent completed with response")
        return Command(
            update={
                "messages": response["messages"],
                "papers": response.get("papers", {}),
                "multi_papers": response.get("multi_papers", {}),
                "last_displayed_papers": response.get("last_displayed_papers", {}),
            },
            # Always return to supervisor
            goto="supervisor",
        )

    # Initialize LLM
    logger.info("Using model %s with temperature %s", llm_model, cfg.temperature)

    # Build the graph
    workflow = StateGraph(Talk2Scholars)
    supervisor = make_supervisor_node(llm_model, thread_id)
    # Add nodes
    workflow.add_node("supervisor", supervisor)
    workflow.add_node("s2_agent", call_s2_agent)
    # Add edges
    workflow.add_edge(START, "supervisor")
    # Compile the workflow
    app = workflow.compile(checkpointer=MemorySaver())
    logger.info("Main agent workflow compiled")
    return app



"""
Tool for performing subgraph summarization.
"""

import logging
from typing import Type, Annotated
from pydantic import BaseModel, Field
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.messages import ToolMessage
from langchain_core.tools.base import InjectedToolCallId
from langchain_core.tools import BaseTool
from langgraph.types import Command
from langgraph.prebuilt import InjectedState
import hydra

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SubgraphSummarizationInput(BaseModel):
    """
    SubgraphSummarizationInput is a Pydantic model representing an input for
    summarizing a given textualized subgraph.

    Args:
        tool_call_id: Tool call ID.
        state: Injected state.
        prompt: Prompt to interact with the backend.
        extraction_name: Name assigned to the subgraph extraction process
    """

    tool_call_id: Annotated[str, InjectedToolCallId] = Field(
        description="Tool call ID."
    )
    state: Annotated[dict, InjectedState] = Field(description="Injected state.")
    prompt: str = Field(description="Prompt to interact with the backend.")
    extraction_name: str = Field(
        description="""Name assigned to the subgraph extraction process
                                 when the subgraph_extraction tool is invoked."""
    )


class SubgraphSummarizationTool(BaseTool):
    """
    This tool performs subgraph summarization over textualized graph to highlight the most
    important information in responding to user's prompt.
    """

    name: str = "subgraph_summarization"
    description: str = """A tool to perform subgraph summarization over textualized graph
                        for responding to user's follow-up prompt(s)."""
    args_schema: Type[BaseModel] = SubgraphSummarizationInput

    def _run(
        self,
        tool_call_id: Annotated[str, InjectedToolCallId],
        state: Annotated[dict, InjectedState],
        prompt: str,
        extraction_name: str,
    ):
        """
        Run the subgraph summarization tool.

        Args:
            tool_call_id: The tool call ID.
            state: The injected state.
            prompt: The prompt to interact with the backend.
            extraction_name: The name assigned to the subgraph extraction process.
        """
        logger.log(
            logging.INFO, "Invoking subgraph_summarization tool for %s", extraction_name
        )

        # Load hydra configuration
        with hydra.initialize(version_base=None, config_path="../configs"):
            cfg = hydra.compose(
                config_name="config", overrides=["tools/subgraph_summarization=default"]
            )
            cfg = cfg.tools.subgraph_summarization

        # Load the extracted graph
        extracted_graph = {dic["name"]: dic for dic in state["dic_extracted_graph"]}
        # logger.log(logging.INFO, "Extracted graph: %s", extracted_graph)

        # Prepare prompt template
        prompt_template = ChatPromptTemplate.from_messages(
            [
                ("system", cfg.prompt_subgraph_summarization),
                ("human", "{input}"),
            ]
        )

        # Prepare chain
        chain = prompt_template | state["llm_model"] | StrOutputParser()

        # Return the subgraph and textualized graph as JSON response
        response = chain.invoke(
            {
                "input": prompt,
                "textualized_subgraph": extracted_graph[extraction_name]["graph_text"],
            }
        )

        # Store the response as graph_summary in the extracted graph
        for key, value in extracted_graph.items():
            if key == extraction_name:
                value["graph_summary"] = response

        # Prepare the dictionary of updated state
        dic_updated_state_for_model = {}
        for key, value in {
            "dic_extracted_graph": list(extracted_graph.values()),
        }.items():
            if value:
                dic_updated_state_for_model[key] = value

        # Return the updated state of the tool
        return Command(
            update=dic_updated_state_for_model
            | {
                # update the message history
                "messages": [ToolMessage(content=response, tool_call_id=tool_call_id)]
            }
        )



'''
Test cases for Talk2Biomodels.
'''

from langchain_core.messages import HumanMessage, ToolMessage
from langchain_openai import ChatOpenAI
from ..agents.t2b_agent import get_app

def test_ask_question_tool():
    '''
    Test the ask_question tool without the simulation results.
    '''
    unique_id = 12345
    app = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}

    ##########################################
    # Test ask_question tool when simulation
    # results are not available i.e. the
    # simulation has not been run. In this
    # case, the tool should return an error
    ##########################################
    # Update state
    app.update_state(config,
            {"llm_model": ChatOpenAI(model='gpt-4o-mini', temperature=0)})
    # Define the prompt
    prompt = "Call the ask_question tool to answer the "
    prompt += "question: What is the concentration of CRP "
    prompt += "in serum at 1000 hours? The simulation name "
    prompt += "is `simulation_name`."
    # Invoke the tool
    app.invoke(
            {"messages": [HumanMessage(content=prompt)]},
            config=config
        )
    # Get the messages from the current state
    # and reverse the order
    current_state = app.get_state(config)
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages until a
    # ToolMessage is found.
    for msg in reversed_messages:
        # Assert that the message is a ToolMessage
        # and its status is "error"
        if isinstance(msg, ToolMessage):
            assert msg.status == "error"



#!/usr/bin/env python3

"""
Tool for asking questions to the article.
"""

import logging
from typing import Type, Annotated
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool
from langchain_core.vectorstores import InMemoryVectorStore
from langchain_community.document_loaders import PyPDFLoader
from langgraph.prebuilt import InjectedState

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class QueryArticleInput(BaseModel):
    """
    Input schema for the query_articles tool.
    """
    question: Annotated[str, Field(description="User question to search articles.")]
    state: Annotated[dict, InjectedState]

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class QueryArticle(BaseTool):
    """
    Tool to ask questions to the article.
    """
    name: str = "query_article"
    description: str = "Ask questions to the article."
    args_schema: Type[BaseModel] = QueryArticleInput

    def _run(self,
             question: str,
             state: Annotated[dict, InjectedState]):
        """
        Run the tool.

        Args:
            query (str): The search query.
        """
        logger.log(logging.INFO, "loading the article from %s", state['pdf_file_name'])
        logger.log(logging.INFO, "searching the article with the question: %s", question)
        # Load the article
        loader = PyPDFLoader(state['pdf_file_name'])
        # Load the pages of the article
        pages = []
        for page in loader.lazy_load():
            pages.append(page)
        # Set up text embedding model
        text_embedding_model = state['text_embedding_model']
        logging.info("Loaded text embedding model %s", text_embedding_model)
        # Create a vector store from the pages
        vector_store = InMemoryVectorStore.from_documents(
                                            pages,
                                            text_embedding_model)
        # Search the article with the question
        docs = vector_store.similarity_search(question)
        # Return the content of the pages
        return "\n".join([doc.page_content for doc in docs])



#!/usr/bin/env python3

"""
Function for loading the BioModel.
"""

from typing import Annotated, Any, Union
from pydantic import BaseModel, BeforeValidator
from ..models.basico_model import BasicoModel

def ensure_biomodel_id(value: Any) -> Any:
    """
    Ensure that the biomodel_id is an integer or a string starting with 'BIOMD' or 'MODEL'.
    """
    if isinstance(value, int):
        return value
    if isinstance(value, str) and (value.startswith("BIOMD") or value.startswith("MODEL")):
        return value
    raise ValueError("biomodel_id must be an integer or a string starting with 'BIOMD' or 'MODEL'.")

class ModelData(BaseModel):
    """
    Base model for the model data.
    """
    biomodel_id: Annotated[Union[int, str], BeforeValidator(ensure_biomodel_id)] = None
    # sbml_file_path: Optional[str] = None
    use_uploaded_sbml_file: bool = False

def load_biomodel(sys_bio_model, sbml_file_path=None):
    """
    Load the BioModel.
    """
    model_object = None
    if sys_bio_model.biomodel_id:
        model_object = BasicoModel(biomodel_id=sys_bio_model.biomodel_id)
    elif sbml_file_path:
        model_object = BasicoModel(sbml_file_path=sbml_file_path)
    return model_object



"""
Test cases for tools/subgraph_extraction.py
"""

import pytest
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from ..agents.t2kg_agent import get_app

# Define the data path
DATA_PATH = "aiagents4pharma/talk2knowledgegraphs/tests/files"


@pytest.fixture(name="input_dict")
def input_dict_fixture():
    """
    Input dictionary fixture.
    """
    input_dict = {
        "llm_model": None,  # TBA for each test case
        "embedding_model": None,  # TBA for each test case
        "uploaded_files": [],
        "topk_nodes": 3,
        "topk_edges": 3,
        "dic_source_graph": [
            {
                "name": "PrimeKG",
                "kg_pyg_path": f"{DATA_PATH}/primekg_ibd_pyg_graph.pkl",
                "kg_text_path": f"{DATA_PATH}/primekg_ibd_text_graph.pkl",
            }
        ],
    }

    return input_dict


def test_extract_subgraph_wo_docs(input_dict):
    """
    Test the subgraph extraction tool without any documents using OpenAI model.

    Args:
        input_dict: Input dictionary.
    """
    # Prepare LLM and embedding model
    input_dict["llm_model"] = ChatOpenAI(model="gpt-4o-mini", temperature=0.0)
    input_dict["embedding_model"] = OpenAIEmbeddings(model="text-embedding-3-small")

    # Setup the app
    unique_id = 12345
    app = get_app(unique_id, llm_model=input_dict["llm_model"])
    config = {"configurable": {"thread_id": unique_id}}
    # Update state
    app.update_state(
        config,
        input_dict,
    )
    prompt = """
    Please directly invoke `subgraph_extraction` tool without calling any other tools 
    to respond to the following prompt:

    Extract all relevant information related to nodes of genes related to inflammatory bowel disease 
    (IBD) that existed in the knowledge graph.
    Please set the extraction name for this process as `subkg_12345`.
    """

    # Test the tool subgraph_extraction
    response = app.invoke({"messages": [HumanMessage(content=prompt)]}, config=config)

    # Check assistant message
    assistant_msg = response["messages"][-1].content
    assert isinstance(assistant_msg, str)

    # Check tool message
    tool_msg = response["messages"][-2]
    assert tool_msg.name == "subgraph_extraction"

    # Check extracted subgraph dictionary
    current_state = app.get_state(config)
    dic_extracted_graph = current_state.values["dic_extracted_graph"][0]
    assert isinstance(dic_extracted_graph, dict)
    assert dic_extracted_graph["name"] == "subkg_12345"
    assert dic_extracted_graph["graph_source"] == "PrimeKG"
    assert dic_extracted_graph["topk_nodes"] == 3
    assert dic_extracted_graph["topk_edges"] == 3
    assert isinstance(dic_extracted_graph["graph_dict"], dict)
    assert len(dic_extracted_graph["graph_dict"]["nodes"]) > 0
    assert len(dic_extracted_graph["graph_dict"]["edges"]) > 0
    assert isinstance(dic_extracted_graph["graph_text"], str)
    # Check if the nodes are in the graph_text
    assert all(
        n[0] in dic_extracted_graph["graph_text"]
        for n in dic_extracted_graph["graph_dict"]["nodes"]
    )
    # Check if the edges are in the graph_text
    assert all(
        ",".join([e[0], '"' + str(tuple(e[2]["relation"])) + '"', e[1]])
        in dic_extracted_graph["graph_text"]
        for e in dic_extracted_graph["graph_dict"]["edges"]
    )


def test_extract_subgraph_w_docs(input_dict):
    """
    Test the subgraph extraction tool with a document as reference (i.e., endotype document)
    using OpenAI model.

    Args:
        input_dict: Input dictionary.
    """
    # Prepare LLM and embedding model
    input_dict["llm_model"] = ChatOpenAI(model="gpt-4o-mini", temperature=0.0)
    input_dict["embedding_model"] = OpenAIEmbeddings(model="text-embedding-3-small")

    # Setup the app
    unique_id = 12345
    app = get_app(unique_id, llm_model=input_dict["llm_model"])
    config = {"configurable": {"thread_id": unique_id}}
    # Update state
    input_dict["uploaded_files"] = [
        {
            "file_name": "DGE_human_Colon_UC-vs-Colon_Control.pdf",
            "file_path": f"{DATA_PATH}/DGE_human_Colon_UC-vs-Colon_Control.pdf",
            "file_type": "endotype",
            "uploaded_by": "VPEUser",
            "uploaded_timestamp": "2024-11-05 00:00:00",
        }
    ]
    app.update_state(
        config,
        input_dict,
    )
    prompt = """
    Please ONLY invoke `subgraph_extraction` tool without calling any other tools 
    to respond to the following prompt:

    Extract all relevant information related to nodes of genes related to inflammatory bowel disease 
    (IBD) that existed in the knowledge graph.
    Please set the extraction name for this process as `subkg_12345`.
    """

    # Test the tool subgraph_extraction
    response = app.invoke({"messages": [HumanMessage(content=prompt)]}, config=config)

    # Check assistant message
    assistant_msg = response["messages"][-1].content
    assert isinstance(assistant_msg, str)

    # Check tool message
    tool_msg = response["messages"][-2]
    assert tool_msg.name == "subgraph_extraction"

    # Check extracted subgraph dictionary
    current_state = app.get_state(config)
    dic_extracted_graph = current_state.values["dic_extracted_graph"][0]
    assert isinstance(dic_extracted_graph, dict)
    assert dic_extracted_graph["name"] == "subkg_12345"
    assert dic_extracted_graph["graph_source"] == "PrimeKG"
    assert dic_extracted_graph["topk_nodes"] == 3
    assert dic_extracted_graph["topk_edges"] == 3
    assert isinstance(dic_extracted_graph["graph_dict"], dict)
    assert len(dic_extracted_graph["graph_dict"]["nodes"]) > 0
    assert len(dic_extracted_graph["graph_dict"]["edges"]) > 0
    assert isinstance(dic_extracted_graph["graph_text"], str)
    # Check if the nodes are in the graph_text
    assert all(
        n[0] in dic_extracted_graph["graph_text"]
        for n in dic_extracted_graph["graph_dict"]["nodes"]
    )
    # Check if the edges are in the graph_text
    assert all(
        ",".join([e[0], '"' + str(tuple(e[2]["relation"])) + '"', e[1]])
        in dic_extracted_graph["graph_text"]
        for e in dic_extracted_graph["graph_dict"]["edges"]
    )



#!/usr/bin/env python3

"""
Enrichment class for enriching PubChem IDs with their STRINGS representation.
"""

import logging
import requests
import hydra

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def drugbank_id2pubchem_cid(drugbank_id):
    """
    Convert DrugBank ID to PubChem CID.

    Args:
        drugbank_id: The DrugBank ID of the drug.

    Returns:
        The PubChem CID of the drug.
    """
    logger.log(logging.INFO, "Load Hydra configuration for PubChem ID conversion.")
    with hydra.initialize(version_base=None, config_path="../configs"):
        cfg = hydra.compose(config_name='config',
                            overrides=['utils/pubchem_utils=default'])
        cfg = cfg.utils.pubchem_utils
    # Prepare the URL
    pubchem_url_for_drug = cfg.drugbank_id_to_pubchem_cid_url + drugbank_id + '/JSON'
    # Get the data
    response = requests.get(pubchem_url_for_drug, timeout=60)
    data = response.json()
    # Extract the PubChem CID
    cid = None
    for substance in data.get("PC_Substances", []):
        for compound in substance.get("compound", []):
            if "id" in compound and "type" in compound["id"] and compound["id"]["type"] == 1:
                cid = compound["id"].get("id", {}).get("cid")
                break
    return cid



'''
Test cases for Talk2Biomodels get_modelinfo tool.
'''

from langchain_core.messages import HumanMessage, ToolMessage
from ..agents.t2b_agent import get_app

def test_get_modelinfo_tool():
    '''
    Test the get_modelinfo tool.
    '''
    unique_id = 12345
    app = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}
    # Update state
    app.update_state(config,
      {"sbml_file_path": ["aiagents4pharma/talk2biomodels/tests/BIOMD0000000449_url.xml"]})
    prompt = "Extract all relevant information from the uploaded model."
    # Test the tool get_modelinfo
    response = app.invoke(
                        {"messages": [HumanMessage(content=prompt)]},
                        config=config
                    )
    assistant_msg = response["messages"][-1].content
    # Check if the assistant message is a string
    assert isinstance(assistant_msg, str)

def test_model_with_no_species():
    '''
    Test the get_modelinfo tool with a model that does not
    return any species.

    This should raise a tool error.
    '''
    unique_id = 12345
    app = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}
    prompt = "Extract all species from model 20"
    # Test the tool get_modelinfo
    app.invoke(
                {"messages": [HumanMessage(content=prompt)]},
                config=config
            )
    current_state = app.get_state(config)
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages until a
    # ToolMessage is found.
    test_condition = False
    for msg in reversed_messages:
        # Check if the message is a ToolMessage from the get_modelinfo tool
        if isinstance(msg, ToolMessage) and msg.name == "get_modelinfo":
            # Check if the message is an error message
            if (msg.status == "error" and
                "ValueError('Unable to extract species from the model.')" in msg.content):
                test_condition = True
                break
    assert test_condition

def test_model_with_no_parameters():
    '''
    Test the get_modelinfo tool with a model that does not
    return any parameters.

    This should raise a tool error.
    '''
    unique_id = 12345
    app = get_app(unique_id)
    config = {"configurable": {"thread_id": unique_id}}
    prompt = "Extract all parameters from model 10"
    # Test the tool get_modelinfo
    app.invoke(
                {"messages": [HumanMessage(content=prompt)]},
                config=config
            )
    current_state = app.get_state(config)
    reversed_messages = current_state.values["messages"][::-1]
    # Loop through the reversed messages until a
    # ToolMessage is found.
    test_condition = False
    for msg in reversed_messages:
        # Check if the message is a ToolMessage from the get_modelinfo tool
        if isinstance(msg, ToolMessage) and msg.name == "get_modelinfo":
            # Check if the message is an error message
            if (msg.status == "error" and
                "ValueError('Unable to extract parameters from the model.')" in msg.content):
                test_condition = True
                break
    assert test_condition



"""
Embedding class using MOLMIM model from NVIDIA NIM.
"""

import json
from typing import List
import requests
from .embeddings import Embeddings

class EmbeddingWithMOLMIM(Embeddings):
    """
    Embedding class using MOLMIM model from NVIDIA NIM
    """
    def __init__(self, base_url: str):
        """
        Initialize the EmbeddingWithMOLMIM class.

        Args:
            base_url: The base URL for the NIM/MOLMIM model.
        """
        # Set base URL
        self.base_url = base_url

    def embed_documents(self, texts: List[str]) -> List[float]:
        """
        Generate embedding for a list of SMILES strings using MOLMIM model.

        Args:
            texts: The list of SMILES strings to be embedded.

        Returns:
            The list of embeddings for the given SMILES strings.
        """
        headers = {
            'accept': 'application/json',
            'Content-Type': 'application/json'
        }
        data = json.dumps({"sequences": texts})
        response = requests.post(self.base_url, headers=headers, data=data, timeout=60)
        embeddings = response.json()["embeddings"]
        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """
        Generate embeddings for an input query using MOLMIM model.

        Args:
            text: A query to be embedded.
        Returns:
            The embeddings for the given query.
        """
        # Generate the embedding
        embeddings = self.embed_documents([text])
        return embeddings



"""
Test cases for datasets/starkqa_primekg_loader.py
"""

import os
import shutil
import pytest
from ..datasets.starkqa_primekg import StarkQAPrimeKG

# Remove the data folder for testing if it exists
LOCAL_DIR = "../data/starkqa_primekg_test/"
shutil.rmtree(LOCAL_DIR, ignore_errors=True)

@pytest.fixture(name="starkqa_primekg")
def starkqa_primekg_fixture():
    """
    Fixture for creating an instance of StarkQAPrimeKGData.
    """
    return StarkQAPrimeKG(local_dir=LOCAL_DIR)

def test_download_starkqa_primekg(starkqa_primekg):
    """
    Test the loading method of the StarkQAPrimeKGLoaderTool class by downloading files
    from HuggingFace Hub.
    """
    # Load StarkQA PrimeKG data
    starkqa_primekg.load_data()
    starkqa_df = starkqa_primekg.get_starkqa()
    primekg_node_info = starkqa_primekg.get_starkqa_node_info()
    split_idx = starkqa_primekg.get_starkqa_split_indicies()
    query_embeddings = starkqa_primekg.get_query_embeddings()
    node_embeddings = starkqa_primekg.get_node_embeddings()

    # Check if the local directory exists
    assert os.path.exists(starkqa_primekg.local_dir)
    # Check if downloaded files exist in the local directory
    files = ['qa/prime/split/test-0.1.index',
             'qa/prime/split/test.index',
             'qa/prime/split/train.index',
             'qa/prime/split/val.index',
             'qa/prime/stark_qa/stark_qa.csv',
             'qa/prime/stark_qa/stark_qa_human_generated_eval.csv',
             'skb/prime/processed.zip']
    for file in files:
        path = f"{starkqa_primekg.local_dir}/{file}"
        assert os.path.exists(path)
    # Check dataframe
    assert starkqa_df is not None
    assert len(starkqa_df) > 0
    assert starkqa_df.shape[0] == 11204
    # Check node information
    assert primekg_node_info is not None
    assert len(primekg_node_info) == 129375
    # Check split indices
    assert list(split_idx.keys()) == ['train', 'val', 'test', 'test-0.1']
    assert len(split_idx['train']) == 6162
    assert len(split_idx['val']) == 2241
    assert len(split_idx['test']) == 2801
    assert len(split_idx['test-0.1']) == 280
    # Check query embeddings
    assert query_embeddings is not None
    assert len(query_embeddings) == 11204
    assert query_embeddings[0].shape[1] == 1536
    # Check node embeddings
    assert node_embeddings is not None
    assert len(node_embeddings) == 129375
    assert node_embeddings[0].shape[1] == 1536

def test_load_existing_starkqa_primekg(starkqa_primekg):
    """

    Test the loading method of the StarkQAPrimeKGLoaderTool class by loading existing files
    in the local directory.
    """
    # Load StarkQA PrimeKG data
    starkqa_primekg.load_data()
    starkqa_df = starkqa_primekg.get_starkqa()
    primekg_node_info = starkqa_primekg.get_starkqa_node_info()
    split_idx = starkqa_primekg.get_starkqa_split_indicies()
    query_embeddings = starkqa_primekg.get_query_embeddings()
    node_embeddings = starkqa_primekg.get_node_embeddings()

    # Check if the local directory exists
    assert os.path.exists(starkqa_primekg.local_dir)
    # Check if downloaded and processed files exist
    files = ['qa/prime/split/test-0.1.index',
             'qa/prime/split/test.index',
             'qa/prime/split/train.index',
             'qa/prime/split/val.index',
             'qa/prime/stark_qa/stark_qa.csv',
             'qa/prime/stark_qa/stark_qa_human_generated_eval.csv',
             'skb/prime/processed.zip']
    for file in files:
        path = f"{starkqa_primekg.local_dir}/{file}"
        assert os.path.exists(path)
    # Check dataframe
    assert starkqa_df is not None
    assert len(starkqa_df) > 0
    assert starkqa_df.shape[0] == 11204
    # Check node information
    assert primekg_node_info is not None
    assert len(primekg_node_info) == 129375
    # Check split indices
    assert list(split_idx.keys()) == ['train', 'val', 'test', 'test-0.1']
    assert len(split_idx['train']) == 6162
    assert len(split_idx['val']) == 2241
    assert len(split_idx['test']) == 2801
    assert len(split_idx['test-0.1']) == 280
    # Check query embeddings
    assert query_embeddings is not None
    assert len(query_embeddings) == 11204
    assert query_embeddings[0].shape[1] == 1536
    # Check node embeddings
    assert node_embeddings is not None
    assert len(node_embeddings) == 129375
    assert node_embeddings[0].shape[1] == 1536



#/usr/bin/env python3

'''
This is the agent file for the Talk2Cells graph.
'''

import logging
import os
from langchain_openai import ChatOpenAI
from langgraph.checkpoint.memory import MemorySaver
from langgraph.graph import START, StateGraph
from langgraph.prebuilt import create_react_agent, ToolNode
from ..tools.scp_agent.search_studies import search_studies
from ..tools.scp_agent.display_studies import display_studies
from ..states.state_talk2cells import Talk2Cells

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def get_app(uniq_id):
    '''
    This function returns the langraph app.
    '''
    def agent_scp_node(state: Talk2Cells):
        '''
        This function calls the model.
        '''
        logger.log(logging.INFO, "Creating SCP_Agent node with thread_id %s", uniq_id)
        # Get the messages from the state
        # messages = state['messages']
        # Call the model
        # inputs = {'messages': messages}
        response = model.invoke(state, {"configurable": {"thread_id": uniq_id}})
        # The response is a list of messages and may contain `tool calls`
        # We return a list, because this will get added to the existing list
        # return {"messages": [response]}
        return response

    # Define the tools
    # tools = [search_studies, display_studies]
    tools = ToolNode([search_studies, display_studies])

    # Create the LLM
    # And bind the tools to it
    # model = ChatOpenAI(model="gpt-4o-mini", temperature=0).bind_tools(tools)

    # Create an environment variable to store the LLM model
    # Check if the environment variable AIAGENTS4PHARMA_LLM_MODEL is set
    # If not, set it to 'gpt-4o-mini'
    llm_model = os.getenv('AIAGENTS4PHARMA_LLM_MODEL', 'gpt-4o-mini')
    # print (f'LLM model: {llm_model}')
    # llm = ChatOpenAI(model="gpt-4o-mini", temperature=0)
    llm = ChatOpenAI(model=llm_model, temperature=0)
    model = create_react_agent(
                            llm,
                            tools=tools,
                            state_schema=Talk2Cells,
                            state_modifier=(
                                            "You are Talk2Cells agent."
                                            ),
                            checkpointer=MemorySaver()
                        )

    # Define a new graph
    workflow = StateGraph(Talk2Cells)

    # Define the two nodes we will cycle between
    workflow.add_node("agent_scp", agent_scp_node)

    # Set the entrypoint as `agent`
    # This means that this node is the first one called
    workflow.add_edge(START, "agent_scp")

    # Initialize memory to persist state between graph runs
    checkpointer = MemorySaver()

    # Finally, we compile it!
    # This compiles it into a LangChain Runnable,
    # meaning you can use it as you would any other runnable.
    # Note that we're (optionally) passing the memory when compiling the graph
    app = workflow.compile(checkpointer=checkpointer)
    logger.log(logging.INFO, "Compiled the graph")

    return app



"""
Test cases for tools/subgraph_summarization.py
"""

import pytest
from langchain_core.messages import HumanMessage
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from ..agents.t2kg_agent import get_app

# Define the data path
DATA_PATH = "aiagents4pharma/talk2knowledgegraphs/tests/files"


@pytest.fixture(name="input_dict")
def input_dict_fixture():
    """
    Input dictionary fixture.
    """
    input_dict = {
        "llm_model": None,  # TBA for each test case
        "embedding_model": None,  # TBA for each test case
        "uploaded_files": [],
        "topk_nodes": 3,
        "topk_edges": 3,
        "dic_source_graph": [
            {
                "name": "PrimeKG",
                "kg_pyg_path": f"{DATA_PATH}/primekg_ibd_pyg_graph.pkl",
                "kg_text_path": f"{DATA_PATH}/primekg_ibd_text_graph.pkl",
            }
        ],
        "dic_extracted_graph": [
            {
                "name": "subkg_12345",
                "tool_call_id": "tool_12345",
                "graph_source": "PrimeKG",
                "topk_nodes": 3,
                "topk_edges": 3,
                "graph_dict": {
                    'nodes': [('IFNG_(3495)', {}), 
                              ('IKBKG_(3672)', {}),
                              ('ATG16L1_(6661)', {}),
                              ('inflammatory bowel disease_(28158)', {}),
                              ('Crohn ileitis and jejunitis_(35814)', {}),
                              ("Crohn's colitis_(83770)", {})],
                    'edges': [('IFNG_(3495)', 'inflammatory bowel disease_(28158)', 
                               {'relation': ['gene/protein', 'associated with', 'disease'],
                                'label': ['gene/protein', 'associated with', 'disease']}), 
                              ('IFNG_(3495)', "Crohn's colitis_(83770)",
                               {'relation': ['gene/protein', 'associated with', 'disease'],
                                'label': ['gene/protein', 'associated with', 'disease']}), 
                              ('IFNG_(3495)', 'Crohn ileitis and jejunitis_(35814)',
                               {'relation': ['gene/protein', 'associated with', 'disease'],
                                'label': ['gene/protein', 'associated with', 'disease']}), 
                              ('ATG16L1_(6661)', 'IKBKG_(3672)',
                               {'relation': ['gene/protein', 'ppi', 'gene/protein'],
                                'label': ['gene/protein', 'ppi', 'gene/protein']}), 
                              ("Crohn's colitis_(83770)", 'ATG16L1_(6661)',
                               {'relation': ['disease', 'associated with', 'gene/protein'],
                                'label': ['disease', 'associated with', 'gene/protein']})]},
                "graph_text": """
            node_id,node_attr
            IFNG_(3495),"IFNG belongs to gene/protein category. 
            This gene encodes a soluble cytokine that is a member of the type II interferon class. 
            The encoded protein is secreted by cells of both the innate and adaptive immune systems. 
            The active protein is a homodimer that binds to the interferon gamma receptor 
            which triggers a cellular response to viral and microbial infections. 
            Mutations in this gene are associated with an increased susceptibility to viral, 
            bacterial and parasitic infections and to several autoimmune diseases. 
            [provided by RefSeq, Dec 2015]."
            IKBKG_(3672),"IKBKG belongs to gene/protein category. This gene encodes the regulatory 
            subunit of the inhibitor of kappaB kinase (IKK) complex, which activates NF-kappaB 
            resulting in activation of genes involved in inflammation, immunity, cell survival, 
            and other pathways. Mutations in this gene result in incontinentia pigmenti, 
            hypohidrotic ectodermal dysplasia, and several other types of immunodeficiencies. 
            A pseudogene highly similar to this locus is located in an adjacent region of the 
            X chromosome. [provided by RefSeq, Mar 2016]."
            ATG16L1_(6661),"ATG16L1 belongs to gene/protein category. The protein encoded 
            by this gene is part of a large protein complex that is necessary for autophagy, 
            the major process by which intracellular components are targeted to lysosomes 
            for degradation. Defects in this gene are a cause of susceptibility to inflammatory 
            bowel disease type 10 (IBD10). Several transcript variants encoding different 
            isoforms have been found for this gene.[provided by RefSeq, Jun 2010]."
            inflammatory bowel disease_(28158),inflammatory bowel disease belongs to disease 
            category. Any inflammatory bowel disease in which the cause of the disease 
            is a mutation in the NOD2 gene.  
            Crohn ileitis and jejunitis_(35814),Crohn ileitis and jejunitis belongs to 
            disease category. An Crohn disease involving a pathogenic inflammatory 
            response in the ileum.  
            Crohn's colitis_(83770),Crohn's colitis belongs to disease category. 
            Crohn's disease affecting the colon.  

            head_id,edge_type,tail_id
            Crohn's colitis_(83770),"('disease', 'associated with', 'gene/protein')",
            ATG16L1_(6661)
            ATG16L1_(6661),"('gene/protein', 'ppi', 'gene/protein')",IKBKG_(3672)
            IFNG_(3495),"('gene/protein', 'associated with', 'disease')",
            inflammatory bowel disease_(28158)
            IFNG_(3495),"('gene/protein', 'associated with', 'disease')",Crohn's colitis_(83770)
            IFNG_(3495),"('gene/protein', 'associated with', 'disease')",
            Crohn ileitis and jejunitis_(35814)
            """,
                "graph_summary": None,
            }
        ],
    }

    return input_dict


def test_summarize_subgraph(input_dict):
    """
    Test the subgraph summarization tool without any documents using Ollama model.

    Args:
        input_dict: Input dictionary fixture.
    """
    # Prepare LLM and embedding model
    input_dict["llm_model"] = ChatOpenAI(model="gpt-4o-mini", temperature=0.0)
    input_dict["embedding_model"] = OpenAIEmbeddings(model="text-embedding-3-small")

    # Setup the app
    unique_id = 12345
    app = get_app(unique_id, llm_model=input_dict["llm_model"])
    config = {"configurable": {"thread_id": unique_id}}
    # Update state
    app.update_state(
        config,
        input_dict,
    )
    prompt = """
    Please directly invoke `subgraph_summarization` tool without calling any other tools 
    to respond to the following prompt:

    You are given a subgraph in the forms of textualized subgraph representing
    nodes and edges (triples) obtained from extraction_name `subkg_12345`.
    Summarize the given subgraph and higlight the importance nodes and edges.
    """

    # Test the tool subgraph_summarization
    response = app.invoke({"messages": [HumanMessage(content=prompt)]}, config=config)

    # Check assistant message
    assistant_msg = response["messages"][-1].content
    assert isinstance(assistant_msg, str)

    # Check tool message
    tool_msg = response["messages"][-2]
    assert tool_msg.name == "subgraph_summarization"

    # Check summarized subgraph
    current_state = app.get_state(config)
    dic_extracted_graph = current_state.values["dic_extracted_graph"][0]
    assert isinstance(dic_extracted_graph["graph_summary"], str)



#!/usr/bin/env python3

"""
This tool is used to return recommendations for a single paper.
"""

import logging
from typing import Annotated, Any, Optional
import hydra
import requests
from langchain_core.messages import ToolMessage
from langchain_core.tools import tool
from langchain_core.tools.base import InjectedToolCallId
from langgraph.types import Command
from pydantic import BaseModel, Field

# Configure logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class SinglePaperRecInput(BaseModel):
    """Input schema for single paper recommendation tool."""

    paper_id: str = Field(
        description="Semantic Scholar Paper ID to get recommendations for (40-character string)"
    )
    limit: int = Field(
        default=5,
        description="Maximum number of recommendations to return",
        ge=1,
        le=500,
    )
    year: Optional[str] = Field(
        default=None,
        description="Year range in format: YYYY for specific year, "
        "YYYY- for papers after year, -YYYY for papers before year, or YYYY:YYYY for range",
    )
    tool_call_id: Annotated[str, InjectedToolCallId]
    model_config = {"arbitrary_types_allowed": True}


# Load hydra configuration
with hydra.initialize(version_base=None, config_path="../../configs"):
    cfg = hydra.compose(
        config_name="config", overrides=["tools/single_paper_recommendation=default"]
    )
    cfg = cfg.tools.single_paper_recommendation


@tool(args_schema=SinglePaperRecInput, parse_docstring=True)
def get_single_paper_recommendations(
    paper_id: str,
    tool_call_id: Annotated[str, InjectedToolCallId],
    limit: int = 5,
    year: Optional[str] = None,
) -> Command[Any]:
    """
    Get recommendations for on a single paper using its Semantic Scholar ID.
    No other ID types are supported.

    Args:
        paper_id (str): The Semantic Scholar Paper ID to get recommendations for.
        tool_call_id (Annotated[str, InjectedToolCallId]): The tool call ID.
        limit (int, optional): The maximum number of recommendations to return. Defaults to 2.
        year (str, optional): Year range for papers.
        Supports formats like "2024-", "-2024", "2024:2025". Defaults to None.

    Returns:
        Dict[str, Any]: The recommendations and related information.
    """
    logger.info(
        "Starting single paper recommendations search with paper ID: %s", paper_id
    )

    endpoint = f"{cfg.api_endpoint}/{paper_id}"
    params = {
        "limit": min(limit, 500),  # Max 500 per API docs
        "fields": ",".join(cfg.api_fields),
        "from": cfg.recommendation_params.from_pool,
    }

    # Add year parameter if provided
    if year:
        params["year"] = year

    response = requests.get(endpoint, params=params, timeout=cfg.request_timeout)
    data = response.json()
    response = requests.get(endpoint, params=params, timeout=10)
    # print(f"API Response Status: {response.status_code}")
    logging.info(
        "API Response Status for recommendations of paper %s: %s",
        paper_id,
        response.status_code,
    )
    if response.status_code != 200:
        raise ValueError("Invalid paper ID or API error.")
    # print(f"Request params: {params}")
    logging.info("Request params: %s", params)

    data = response.json()
    recommendations = data.get("recommendedPapers", [])

    if not recommendations:
        return Command(
            update={
                "papers": {},
                "messages": [
                    ToolMessage(
                        content=f"No recommendations found for {paper_id}.",
                        tool_call_id=tool_call_id,
                    )
                ],
            }
        )

    # Extract paper ID and title from recommendations
    filtered_papers = {
        paper["paperId"]: {
            # "semantic_scholar_id": paper["paperId"],  # Store Semantic Scholar ID
            "Title": paper.get("title", "N/A"),
            "Abstract": paper.get("abstract", "N/A"),
            "Year": paper.get("year", "N/A"),
            "Citation Count": paper.get("citationCount", "N/A"),
            "URL": paper.get("url", "N/A"),
            # "arXiv_ID": paper.get("externalIds", {}).get(
            #     "ArXiv", "N/A"
            # ),  # Extract arXiv ID
        }
        for paper in recommendations
        if paper.get("title") and paper.get("authors")
    }

    content = "Recommendations based on a single paper were successful."
    content += " Here is a summary of the recommendations:"
    content += f"Number of papers found: {len(filtered_papers)}\n"
    content += f"Query Paper ID: {paper_id}\n"
    content += f"Year: {year}\n" if year else ""

    return Command(
        update={
            "papers": filtered_papers,  # Now sending the dictionary directly
            "last_displayed_papers": "papers",
            "messages": [
                ToolMessage(
                    content=content,
                    tool_call_id=tool_call_id,
                    artifact=filtered_papers,
                )
            ],
        }
    )
